﻿/*
Copyright (c) 2003-2011, CKSource - Frederico Knabben. All rights reserved.
For licensing, see LICENSE.html or http://ckeditor.com/license
 */

(function() {
	CKEDITOR.plugins.add('tracwiki', {
		requires : [ 'htmlwriter' ],

		init : function(editor) {
			var dataProcessor = editor.dataProcessor = new CKEDITOR.tracwikiDataProcessor(editor);

			dataProcessor.writer.forceSimpleAmpersand = editor.config.forceSimpleAmpersand;
		},

		onLoad : function() {
			!('fillEmptyBlocks' in CKEDITOR.config) && (CKEDITOR.config.fillEmptyBlocks = 1);
		}
	});

	CKEDITOR.tracwikiDataProcessor = function(editor) {
		this.editor = editor;

		this.writer = new CKEDITOR.htmlWriter();
		this.dataFilter = new CKEDITOR.htmlParser.filter();
		this.htmlFilter = new CKEDITOR.htmlParser.filter();
	};

	CKEDITOR.tracwikiDataProcessor.prototype = {
		toHtml : function(data, fixForBody) {
			// @todo: Display a "load in progress" message somewhere

			// Transform TracWiki to HTML by requesting Trac server to do it
			// using AJAX.
			var ajax_data = {};
			ajax_data["realm"] = ck_resource_realm;
			ajax_data["id"] = ck_resource_id;
			ajax_data["__FORM_TOKEN"] = form_token;
			ajax_data["text"] = data;
			var ajaxResponse = jQuery.ajax({
				type : "POST",
				url : ck_render_url,
				data : ajax_data,
				dataType : "html",
				async : false
			// success: this.ajaxSuccess, error: this.ajaxError
			});
			// @todo: Error handling for AJAX request
			data = ajaxResponse.responseText;
			return data;
		},

		toDataFormat : function(html, fixForBody) {
			var list_str, offset = '', tablewidth, tableheight;
			var hpattern = /^h\d/;
			var intablepattern = /t([d|h|r|body|head|able])/;
			var inlistpattern = /([ol|ul])/;
			var widthpattern = /width:\s(\d+)(px|%);/;
			var heightpattern = /height:\s(\d+)(px|%);/;
			var marginpattern = /margin-left:\s(\d+)px;/;
			var urlpattern = /http:\/\/.*\/([ticket\/\d+|wiki\/.*|htdocs])\/(.*\.\B[gif|jpg|png])/;
			var formatters = [ "'''", "''", '__', '`', '~~', '^', ',,' ];

			if ('html_wrapper' == ck_editor_type) {
				return '{{{\n#!html\n' + html + '\n}}}';
			} else {
				var writer = this.writer;

				this.getHtmlBlock = function(fragment) {
					writer.reset();
					fragment.writeHtml(writer, this.htmlFilter);
					return '[[html(' + writer.getHtml(true) + ')]]';
				};

				// @todo: Handle HTML->TracWiki conversion in an extensible
				// way...
				this.parseFragment = function(fragment) {

					var data = '', frag, style;
					var marginoffset;

					for ( var i = 0; i < fragment.children.length; i++) {
						frag = fragment.children[i];
						style = null;
						marginoffset = '';
						if (frag.name) {
							if (frag.attributes.style && frag.attributes.style != null) {
								style = frag.attributes.style;
								// margins for normal text
								var margin = style.match(marginpattern);
								if (margin != null) {
									for ( var j = 1; j < margin[1]; j += 40)
										marginoffset += ' ';
									style = style.replace(margin[0], '');
									style = (style == '') ? null : style;
								}
							}
							if (style != null && frag.name != 'table' && frag.name != 'th' && frag.name != 'td') {
								data += this.getHtmlBlock(frag);
							} else {
								var subcontent;
								switch (frag.name) {
								// Headlines
								case 'h1':
									data += '\n' + offset + '= ' + this.parseFragment(frag) + ' = \n';
									break;
								case 'h2':
									data += '\n' + offset + '== ' + this.parseFragment(frag) + ' == \n';
									break;
								case 'h3':
									data += '\n' + offset + '=== ' + this.parseFragment(frag) + ' === \n';
									break;
								case 'h4':
									data += '\n' + offset + '==== ' + this.parseFragment(frag) + ' ==== \n';
									break;
								case 'h5':
									data += '\n' + offset + '===== ' + this.parseFragment(frag) + ' ===== \n';
									break;
								case 'h6':
									data += '\n' + offset + '====== ' + this.parseFragment(frag) + ' ====== \n';
									break;

								// Textformat
								case 'u':
									data += '__' + this.parseFragment(frag) + '__';
									break;
								case 'strong':
									data += "'''" + this.parseFragment(frag) + "'''";
									break;
								case 'em':
									data += "''" + this.parseFragment(frag) + "''";
									break;
								case 'sup':
									data += '^' + this.parseFragment(frag) + '^';
									break;
								case 'sub':
									data += ',,' + this.parseFragment(frag) + ',,';
									break;
								case 'strike':
									data += '~~' + this.parseFragment(frag) + '~~';
									break;
								case 'del':
									data += '~~' + this.parseFragment(frag) + '~~';
									break;
								case 'tt':
									data += '`' + this.parseFragment(frag) + '`';
									break;
								case 'hr': // horizontal row
									data += '\n----\n';
									break;
								case 'a': // hyperlinks
									subcontent = this.parseFragment(frag);
									if (frag.attributes.rel.match('nofollow') != null)
										subcontent = subcontent.replace("\?", "");
									data += subcontent;
									break;
								case 'img': // hyperlinks
									var src = frag.attributes.src.match(urlpattern);
									if (src != null)
										data += ' [[Image(ticket:' + src[1] + ':' + src[2] + ')]]';
									else
										data += this.getHtmlBlock(frag);
									break;

								// Container
								case 'p':
									var linebreaks = true;
									subcontent = this.parseFragment(frag);
									var par = frag.parent;
									while (par && par.name) {
										if (par.name.match("td|th|li") != null) {
											linebreaks = false;
											break;
										}
										par = parent.parent;
									}
									if (linebreaks)
										data += offset + marginoffset + subcontent + "\n\n";
									else
										data += subcontent;
									break;
								case 'blockquote':
									data += this.parseFragment(frag);
									break;
								case 'div':
									data += this.parseFragment(frag);
									break;
								case 'pre':
									data += '\n{{{\n';
									var c = this.parseFragment(frag);
									c = c.replace(/\n*$/, "");
									data += c;
									data += '\n}}}\n';
									break;
								case 'span':
									if (frag.className == 'underline')
										data += '__' + this.parseFragment(frag) + '__';
									else
										data += this.parseFragment(frag);
									break;

								case 'ul': // unordered List
									list_str = '- ';
									offset += ' ';
									data += this.parseFragment(frag)
									if (!frag.parent || !frag.parent.name || !frag.parent.name.match(inlistpattern))
										data += '\n\n';
									offset = offset.replace(/^\s/, '');
									break;
								case 'ol': // ordered list
									if (frag.attributes.start)
										list_str = frag.attributes.start + '. ';
									else
										list_str = '1. ';
									offset += ' ';
									data += this.parseFragment(frag)
									if (!frag.parent || !frag.parent.name || !frag.parent.name.match(inlistpattern))
										data += '\n\n';
									offset = offset.replace(/^\s/, '');
									break;
								case 'li':
									var subcontent = this.parseFragment(frag);
									data += '\n' + marginoffset + offset + list_str + subcontent;
									break;

								case 'br': // line break
									data += ' [[BR]]';
									if (frag.parent && frag.parent.name) {
										if (!frag.parent.name.match(intablepattern) && !frag.parent.name.match(hpattern) && !frag.parent.name == 'li') {
											data += '\n';
											break;
										}
									}
									break;

								case 'table': // table
									offset += ' ';
									if (frag.attributes.style && frag.attributes.style != null) {
										tablewidth = frag.attributes.style.match(widthpattern);
										tableheight = frag.attributes.style.match(heightpattern);
									}
									data += this.parseFragment(frag).replace(/\|---------------$/, '\n');
									offset = offset.replace(/^\s/, '');
									break;
								case 'tbody':
									data += this.parseFragment(frag);
									break;
								case 'thead':
									data += this.parseFragment(frag);
									break;
								// Complex Tables
								case 'tr':
									data += this.parseFragment(frag);
									if (frag.parent.name != 'tr')
										data += '\n' + offset + '|---------------';
									break;
								case 'th':
									data += '\n' + offset + '{{{#!th' + this.getTableCellAttributes(frag);
									data += '\n' + this.parseFragment(frag).replace("\n*$", "");
									data += '\n' + offset + '}}}';
									break;
								case 'td':
									data += '\n' + offset + '{{{#!td' + this.getTableCellAttributes(frag);
									data += '\n' + this.parseFragment(frag).replace("\n*$", "");
									data += '\n' + offset + '}}}';
									break;

								default: // unknown/ unsupported formats
									data += this.getHtmlBlock(frag);
								}
							}

						} else {
							var value = frag.value;
							value = value.replace(/&nbsp;/g, ' ');
							value = value.replace(/&quot;/g, '\"');
							value = value.replace(/&amp;/g, '\&');
							value = value.replace(/&lt;/g, '\<');
							value = value.replace(/&gt;/g, '\>');
							data += value;
						}
					}
					if ('!' == data.substr(-1))
						data += '';
					return data;
				}
			}
			;

			this.getTableCellAttributes = function(frag) {
				var attrib = '';
				attrib += ' style="' + this.getTableFieldWidth(frag) + this.getTableFieldHeight(frag) + 'background: #' + (frag.name == 'th' ? 'dde' : 'eef')
						+ '"';
				attrib += frag.attributes.colspan ? ' colspan=' + frag.attributes.colspan : '';
				attrib += frag.attributes.rowspan ? ' rowspan=' + frag.attributes.rowspan : '';
				attrib += frag.attributes.scope ? ' scope=' + frag.attributes.scope : '';
				attrib += frag.attributes.align ? ' align=' + frag.attributes.align : ' align=justify';
				attrib = attrib.replace(/undefined/g, '');
				return attrib;
			};

			this.getTableFieldHeight = function(frag) {
				var height = null;
				if (frag.attributes.style && frag.attributes.style != null)
					height = frag.attributes.style.match(heightpattern);
				if (height != null)
					return height[0];
				else {
					var columncount = 0, row = frag, table = frag;
					while (row.name != 'tr' && row.parent.name.match(intablepattern) != null)
						row = row.parent;
					if (tableheight != null) {
						for ( var i = 0; i < row.children.length; i++) {
							var span = row.children[i].attributes.colspan ? row.children[i].attributes.colspan : 1;
							columncount += row.children[i].name == 'br' ? 0 : span;
						}
						var ownspan = frag.attributes.colspan ? frag.attributes.colspan : 1;
						height = 'height: ' + ((tableheight[1] / columncount) * ownspan) + '' + tableheight[2] + '; ';
					}
				}
				return (height == null) ? '' : height;
			};

			this.getTableFieldWidth = function(frag) {
				var width = null;
				if (frag.attributes.style && frag.attributes.style != null)
					width = frag.attributes.style.match(widthpattern);
				if (width != null)
					return width[0];
				else {
					var columncount = 0, row = frag, table = frag;
					while (row.name != 'tr' && row.parent.name.match(intablepattern) != null)
						row = row.parent;
					if (tablewidth != null) {
						for ( var i = 0; i < row.children.length; i++) {
							var span = row.children[i].attributes.colspan ? row.children[i].attributes.colspan : 1;
							columncount += row.children[i].name == 'br' ? 0 : span;
						}
						var ownspan = frag.attributes.colspan ? frag.attributes.colspan : 1;
						width = 'width: ' + ((tablewidth[1] / columncount) * ownspan) + '' + tablewidth[2] + '; ';
					}
				}
				return (width == null) ? '' : width;
			};

			this.cleanWiki = function(content) {
				// for ( var i = 0; i < formatters.length; i++) {
				// var f = new RegExp(formatters[i] + formatters[i], "g");
				// content = content.replace(f, "");
				// }
				content = content.replace(/\n((\n+)" + \[\[BR\]\] + "\n)*\n/, "\n\n");
				content = content.replace(/\n\n+/, "\n\n");
				content = content.replace(/"(" + \[\[BR\]\] + "\n)*\\z/, "");
				return content;
			};

			var fragment = CKEDITOR.htmlParser.fragment.fromHtml(html, fixForBody);
			var wiki = this.parseFragment(fragment)
			wiki = this.cleanWiki(wiki);
			return wiki;
		}
	};
})();
